/* acconfig.h
   This file is in the public domain.

   Descriptive text for the C preprocessor macros that
   the distributed Autoconf macros can define.
   No software package will use all of them; autoheader copies the ones
   your configure.in uses into your configuration header file templates.

   The entries are in sort -df order: alphabetical, case insensitive,
   ignoring punctuation (such as underscores).  Although this order
   can split up related entries, it makes it easier to check whether
   a given entry is in the file.

   Leave the following blank line there!!  Autoheader needs it.  */
^L

#undef APPDEF		/* The full app-defaults file path.  */

#undef HAVE_GETHOSTNAME

#undef HAVE_LIBCURSES
#undef HAVE_LIBX11

#undef HAVE_NCURSES_H

#undef MAPPATH		/* Systemwide path for maps.  */

#undef VERSION_STR	/* Version string in the form "motti x.x".
			   Best defined at compile-time since it's
			   fixed for the executable.  */

^L
/* Leave that blank line there!!  Autoheader needs it.
   If you're adding to this file, keep in mind:
   The entries are in sort -df order: alphabetical, case insensitive,
   ignoring punctuation (such as underscores).  */
